/* WSPacket -- Wrapper around data kept by Burp for each websockets message.
 *
 * Copyright (c) 2014 Eddy Vanlerberghe.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of Eddy Vanlerberghe shall not be used to endorse or promote
 *    products derived from this software without specific prior written
 *    permission.
 *
 * THIS SOFTWARE IS PROVIDED BY EDDY VANLERBERGHE ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL EDDY VANLERBERGHE BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
package evilwan;

import java.util.*;
import java.lang.reflect.*;

public class WSPacket {
    // {{ Class members }}

    /**
     * Burp websocket message encapsulation class
     */
    private static Class	_wsclazz = null;
    /**
     * Burp byte[] encapsulation class
     */
    private static Class	_barclazz = null;
    /**
     * Burp URL encapsulation class
     */
    private static Class	_earlclazz = null;
    /**
     * Accessor method on _barclazz instances
     */
    private static Method	_burp2bar = null;
    /**
     * Accessor methods on _wsclazz instances
     */
    private static Method	_get_comment = null;
    private static Method	_get_time = null;
    private static Method	_get_c2s = null;
    private static Method	_get_url = null;
    private static Method	_get_orgmsg = null;
    private static Method	_get_cgdmsg = null;
    
    // {{ Instance members }}
    
    /**
     * Encapsulated Burp websocket message object
     */
    private Object _burp_packet = null;

    // {{ Class methods }}


    //
    // Analyze constructor for _wsclazz and return burp specific class of which there
    // are two arguments in the constructor: that is the byte[] encapsulation class we need.
    // Apparently there is also an URL encapsulating class in Burp: that type occurs exactly
    // once in the constructor arguments.
    //
    private static void getByteArrayEncapsulation(Class clz) throws Exception {
	Constructor[] inits = clz.getDeclaredConstructors();
	if(inits.length != 1) {
	    throw new Exception("class " + clz.getName() + " expected exactly one constructor, found " +
				inits.length);
	}
	Class[] paramtypes = inits[0].getParameterTypes();
	Hashtable<Class,Integer> ht = new Hashtable<Class,Integer>();
	for(int i = 0; i < paramtypes.length; ++i) {
	    if(paramtypes[i].getName().startsWith("burp")) {
		//
		// Burp specific class: can be an byte[] or a URL encapsulation
		//
		Integer n = ht.get(paramtypes[i]);
		if(n == null) {
		    n = new Integer(0);
		}
		n = new Integer(n.intValue() + 1);
		ht.put(paramtypes[i], n);
	    }
	}
	for(Enumeration<Class> e = ht.keys(); e.hasMoreElements(); ) {
	    Class cc = e.nextElement();
	    if(ht.get(cc).intValue() == 2) {
		_barclazz = cc;	// byte[] encapsulation occurs twice in constructor argument list
	    } else if(ht.get(cc).intValue() == 1) {
		_earlclazz = cc;	// URL encapsulation occurs once in constructor argument list
	    }
	}
    }

    private static Method getBarExtraction(Class c) {
	Method m = null;
	Method[] mets = c.getDeclaredMethods();
	for(int i = 0; i < mets.length; ++i) {
	    Class[] paramtypes = mets[i].getParameterTypes();
	    if(paramtypes.length == 0) {
		if(mets[i].getReturnType().equals(byte[].class)) {
		    m = mets[i];
		    break;
		}
	    }
	}
	return m;
    }
    //
    // Code based on analysis of the actual ArrayList element Burp class.
    //
    // First step in that analysis is to dump the full class name for those array elements ("burp.syc"). Next step
    // is to use javap to dump the contents of that class (including constant pool entries and bytecode)
    // The first interesting thing to notice in the javap output is that several UTF8 entries in the constant pool
    // contain apparent random data instead of regular string constants. A small test using Java reflection that dumped
    // all "static String" and "String[]" fields from class "burp.syc" reveals that a "private static String[]"
    // with name "z" contains some interesting values such as "clientToServer", "originalMessage" etc.
    //
    // The bytecode for method "public void a(burp.dec)" contains many references to sequences like this:
    //
    //	getstatic     #266                // Field z:[Ljava/lang/String;
    //	iconst_2      			  // pushes integer value 2 to stack
    //	aaload        			  // pushes reference to "z[2]" to stack
    //	aload_0       
    //	getfield      #7                  // Field h:Lburp/ckb;
    //	invokeinterface #31,  3           // InterfaceMethod burp/dec.a:(Ljava/lang/String;Lburp/ckb;)V
    //
    // or
    //
    //	getstatic     #266                // Field z:[Ljava/lang/String;
    //	bipush        7			  // pushes integer value 7 to stack
    //	aaload        			  // pushes reference to "z[7]" to stack
    //	aload_0       
    //	getfield      #8                  // Field l:Lburp/ckb;
    //	invokeinterface #31,  3           // InterfaceMethod burp/dec.a:(Ljava/lang/String;Lburp/ckb;)V
    //
    // So basically references to entries in the String[] "z" that was found by dumping the class String constants.
    //
    // For the two examples above, these two refer to "originalMessage" and "autoModifiedMessage" respectively.
    //
    // Anyway, the sequence of calls with each time a String containing something that we would expect to be
    // a property for a generalized websocket packet object and an instance property ("h" and "l" in the
    // examples, both of type "burp.ckb") indicates that we found a link between the internal fields and
    // their description. Further experimentation confirmed this.
    //
    // The next step in the analysis is to note that class "burp.syc" has accessor methods for its properties:
    // those methods are what we need for our dumping purpose.
    //
    // Wherever possible, the code below uses general reasons for determinging the meaning of the individual
    // properties, but where this is not possible (e.g. there are 3 properties that contain message data) the names
    // that were found by means of linking strings to field references in javap output is used.
    // It should be noted that all of those will probably change with each new build of Burp.
    //
    // For now it is unknown what the meaning is of the automodified version of the message, wo we ignore
    // that data in the output. Equally so, highlighting information is ignored for now.
    //
    private static void getAccessorMethods(Class c) {
	Method[] mets = c.getDeclaredMethods();
	for(int i = 0; i < mets.length; ++i) {
	    Class[] paramtypes = mets[i].getParameterTypes();
	    if(paramtypes.length == 0) {
		if(mets[i].getReturnType().equals(long.class)) {
		    //
		    // long is typically used for timestamps and there is only one long property
		    //
		    _get_time = mets[i];
		} else if(mets[i].getReturnType().equals(boolean.class)) {
		    //
		    // boolean is typically used for items that can have only two values and there is
		    // only one boolean property
		    //
		    _get_c2s = mets[i];
		} else if(mets[i].getReturnType().equals(_barclazz) && mets[i].getName().equals("g")) {
		    _get_orgmsg = mets[i];
		} else if(mets[i].getReturnType().equals(String.class) && mets[i].getName().equals("b")) {
		    _get_comment = mets[i];
		} else if(mets[i].getReturnType().equals(_barclazz) && mets[i].getName().equals("f")) {
		    _get_cgdmsg = mets[i];
		} else if(mets[i].getReturnType().equals(_earlclazz)) {
		    //
		    // Needs further refinement for getting actual URL string: for time being just call "toString()"
		    // on whatever that is returned by this method
		    //
		    _get_url = mets[i];
		}
	    }
	}
    }

    //
    // Get remaining info from Burp environment for working with websocket messages.
    //
    // This includes extracting accessor methods for various websocket message properties, as well as
    // accessor method for extracting the raw message bytes (it appears that the TableModel ArrayList
    // does not store byte[] directly but instead uses another (or more?) encapsulation layer(s)
    //
    // For this remaining analysis we start by using info that can be obtained from the first element
    // in the ArrayList (such as class and accessor methods)
    //
    // After executing this method we have all the info we need to dump the websocket messages.
    //
    public static void init(Object datalist) {
	try {
	    ArrayList al = (ArrayList) datalist;
	    if(al.size() <= 0) {
		throw new Exception("WebSocket table data model is empty.");
	    }
	    Object o = al.get(0);	// get first element from that list for further investigation
	    _wsclazz = o.getClass();	// = "burp.syc" (in version 1.6)
	    getByteArrayEncapsulation(_wsclazz);	// _barclazz = "burp.ckb", _earlclazz = "burp.iib" (in version 1.6)
	    getAccessorMethods(_wsclazz);
	    _burp2bar = getBarExtraction(_barclazz);	// accessor method for extracting actual message bytes
	} catch(Throwable t) {
	    System.out.println("## caught: " + t);
	    t.printStackTrace();
	    System.exit(1);
	}
    }

    // {{ Constructors }}
    
    /**
     * Default public constructor.
     */
    public WSPacket(Object bp) {
	_burp_packet = bp;
    }
    
    // {{ Instance methods }}

    /**
     * Accessor methods to Burp websocket message objects
     */
    public boolean getClientToServer() {
	try {
	    return ((Boolean)_get_c2s.invoke(_burp_packet)).booleanValue();
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return false;
    }
    public long getTimestamp() {
	try {
	    return ((Long)_get_time.invoke(_burp_packet)).longValue();
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return 0L;
    }
    public String getComment() {
	try {
	    return (String)_get_comment.invoke(_burp_packet);
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
    public String getURL() {
	try {
	    Object o = _get_url.invoke(_burp_packet);;
	    return _get_url.invoke(_burp_packet).toString();
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
    public byte[] getOriginalMessage() {
	try {
	    Object o = _get_orgmsg.invoke(_burp_packet);;
	    if(o == null) {
		return null;
	    } else {
		return (byte[])_burp2bar.invoke(o);
	    }
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
    public byte[] getEditedMessage() {
	try {
	    Object o = _get_cgdmsg.invoke(_burp_packet);;
	    if(o == null) {
		return null;
	    } else {
		return (byte[])_burp2bar.invoke(o);
	    }
	} catch(Exception iae) {
	    // NOP: "should never happen" ;-)
	}
	return null;
    }
}
